import {
  ClientSideBasePluginConfig,
  ClientSideBaseVisitor,
  DocumentMode,
  getConfigValue,
  indentMultiline,
  LoadedFragment,
} from '@graphql-codegen/visitor-plugin-common';
import autoBind from 'auto-bind';
import { GraphQLSchema, Kind, OperationDefinitionNode, print } from 'graphql';
import { RawGraphQLRequestPluginConfig } from './config';

export interface GraphQLRequestPluginConfig extends ClientSideBasePluginConfig {
  rawRequest: boolean;
  extensionsType: string;
  useSubscriptionClient: boolean;
}

const additionalExportedTypes = `
export type SdkFunctionWrapper = <T>(action: (requestHeaders?:Record<string, string>) => Promise<T>, operationName: string, operationType?: string) => Promise<T>;
`;

export class GraphQLRequestVisitor extends ClientSideBaseVisitor<
  RawGraphQLRequestPluginConfig,
  GraphQLRequestPluginConfig
> {
  private _operationsToInclude: {
    node: OperationDefinitionNode;
    documentVariableName: string;
    operationType: string;
    operationResultType: string;
    operationVariablesTypes: string;
  }[] = [];

  constructor(schema: GraphQLSchema, fragments: LoadedFragment[], rawConfig: RawGraphQLRequestPluginConfig) {
    super(schema, fragments, rawConfig, {
      rawRequest: getConfigValue(rawConfig.rawRequest, false),
      extensionsType: getConfigValue(rawConfig.extensionsType, 'any'),
      useSubscriptionClient: getConfigValue(rawConfig.useSubscriptionClient, false),
    });

    autoBind(this);

    const typeImport = this.config.useTypeImports ? 'import type' : 'import';
    if (this.config.useSubscriptionClient)
      this._additionalImports.push(
        `${typeImport} { GraphQLSubscriptionClient, GraphQLSubscriber, UnsubscribeCallback } from 'graphql-request/dist/types';`
      );
    else this._additionalImports.push(`${typeImport} { GraphQLClient } from 'graphql-request';`);
    this._additionalImports.push(`${typeImport} * as Dom from 'graphql-request/dist/types.dom';`);

    if (this.config.rawRequest) {
      this._additionalImports.push(`${typeImport} { GraphQLError } from 'graphql-request/dist/types';`);
      if (this.config.documentMode !== DocumentMode.string) {
        this._additionalImports.push(`import { print } from 'graphql'`);
      }
    }
  }

  public OperationDefinition(node: OperationDefinitionNode) {
    const operationName = node.name?.value;

    if (!operationName) {
      // eslint-disable-next-line no-console
      console.warn(
        `Anonymous GraphQL operation was ignored in "typescript-graphql-request", please make sure to name your operation: `,
        print(node)
      );

      return null;
    }

    return super.OperationDefinition(node);
  }

  protected buildOperation(
    node: OperationDefinitionNode,
    documentVariableName: string,
    operationType: string,
    operationResultType: string,
    operationVariablesTypes: string
  ): string {
    this._operationsToInclude.push({
      node,
      documentVariableName,
      operationType,
      operationResultType,
      operationVariablesTypes,
    });

    return null;
  }

  private getDocumentNodeVariable(documentVariableName: string): string {
    return this.config.documentMode === DocumentMode.external
      ? `Operations.${documentVariableName}`
      : documentVariableName;
  }

  public get sdkContent(): string {
    const extraVariables: string[] = [];
    const allPossibleActions = this._operationsToInclude
      .map(o => {
        const operationType = o.node.operation;
        const operationName = o.node.name.value;
        const optionalVariables =
          !o.node.variableDefinitions ||
          o.node.variableDefinitions.length === 0 ||
          o.node.variableDefinitions.every(v => v.type.kind !== Kind.NON_NULL_TYPE || v.defaultValue);
        const docVarName = this.getDocumentNodeVariable(o.documentVariableName);
        let docArg = undefined;
        if (this.config.rawRequest) {
          docArg = docVarName;
          if (this.config.documentMode !== DocumentMode.string) {
            docArg = `${docVarName}String`;
            extraVariables.push(`const ${docArg} = print(${docVarName});`);
          }
        }
        if (o.node.operation === 'subscription') {
          if (!this.config.useSubscriptionClient) return '';
          if (this.config.rawRequest) {
            return `${operationName}(${
              optionalVariables ? '' : 'variables: ' + o.operationVariablesTypes + ', '
            }subscriber: GraphQLSubscriber<${o.operationResultType}>): UnsubscribeCallback {
    return client.rawSubscribe(${docArg}, subscriber${optionalVariables ? '' : ', variables'});
}`;
          } else {
            return `${operationName}(${
              optionalVariables ? '' : 'variables: ' + o.operationVariablesTypes + ', '
            }subscriber: GraphQLSubscriber<${o.operationResultType}>): UnsubscribeCallback {
    return client.subscribe(${docVarName}, subscriber${optionalVariables ? '' : ', variables'});
}`;
          }
        } else if (this.config.rawRequest) {
          return `${operationName}(variables${optionalVariables ? '?' : ''}: ${
            o.operationVariablesTypes
          }, requestHeaders?: Dom.RequestInit["headers"]): Promise<{ data?: ${
            o.operationResultType
          } | undefined; extensions?: ${this.config.extensionsType}; ${
            this.config.useSubscriptionClient ? '' : 'headers:  Dom.Headers; status: number; '
          }errors?: GraphQLError[] | undefined; }> {
    return withWrapper((wrappedRequestHeaders) => client.rawRequest<${
      o.operationResultType
    }>(${docArg}, variables, {...requestHeaders, ...wrappedRequestHeaders}), '${operationName}', '${operationType}');
}`;
        } else {
          return `${operationName}(variables${optionalVariables ? '?' : ''}: ${
            o.operationVariablesTypes
          }, requestHeaders?: Dom.RequestInit["headers"]): Promise<${o.operationResultType}> {
  return withWrapper((wrappedRequestHeaders) => client.request<${
    o.operationResultType
  }>(${docVarName}, variables, {...requestHeaders, ...wrappedRequestHeaders}), '${operationName}', '${operationType}');
}`;
        }
      })
      .filter(Boolean)
      .map(s => indentMultiline(s, 2));

    return `${additionalExportedTypes}

const defaultWrapper: SdkFunctionWrapper = (action, _operationName, _operationType) => action();
${extraVariables.join('\n')}
export function getSdk(client: ${
      this.config.useSubscriptionClient ? 'GraphQLSubscriptionClient' : 'GraphQLClient'
    }, withWrapper: SdkFunctionWrapper = defaultWrapper) {
  return {
${allPossibleActions.join(',\n')}
  };
}
export type Sdk = ReturnType<typeof getSdk>;`;
  }
}
